package com.prosc.commerce.gateway.linkpoint;

import com.prosc.commerce.gateway.*;
import com.prosc.commerce.gateway.linkpoint.input.FirstDataOrder;
import com.prosc.commerce.gateway.linkpoint.input.FirstDataPreviousOrderIdentifier;
import com.prosc.commerce.gateway.linkpoint.ws.*;
import com.prosc.commerce.input.*;
import com.prosc.ws.SOAPLoggingHandler;

import javax.xml.namespace.QName;
import javax.xml.ws.Binding;
import javax.xml.ws.BindingProvider;
import javax.xml.ws.handler.Handler;
import javax.xml.ws.soap.SOAPFaultException;
import java.math.BigInteger;
import java.net.URL;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Created by IntelliJ IDEA.
 * User: sterling
 * Date: 3/6/12
 * Time: 1:15 PM
 * To change this template use File | Settings | File Templates.
 */
public class FirstDataGlobalGateway extends Gateway {
	private static final Logger log = Logger.getLogger(FirstDataGlobalGateway.class.getName());

	private static final DateFormat monthFormat = new SimpleDateFormat("MM");
	private static final DateFormat yearFormat = new SimpleDateFormat("yy");
	private static final DateFormat recurringFormat = new SimpleDateFormat("yyyyMMdd");

	private static final String PRODUCTION_URL = "https://ws.firstdataglobalgateway.com/fdggwsapi/services";
	private static final String TEST_URL = "https://ws.merchanttest.firstdataglobalgateway.com/fdggwsapi/services";

	private static final String MODIFY_FUNCTION = "modify";
	private static final String CANCEL_FUNCTION = "cancel";

	private static FDGGWSApiOrderService service;
	private SOAPLoggingHandler loggingHandler;
	private SoapFaultEditingHandler soapFaultEditingHandler;

	private FDGGWSApiOrderResponse orderResponse;
	private FDGGWSApiActionResponse actionResponse;

	private TransactionTypeEnum transactionType = TransactionTypeEnum.AuthCaptureTxn;

	public static enum AccountType {
		PRIMARY_CHECKING("PC"), PRIMARY_SAVINGS("PS"), BACKUP_CHECKING("BC"), BACKUP_SAVINGS("BS");

		String accountTypeAbbreviation;

		AccountType(String accountTypeAbbreviation) {
			this.accountTypeAbbreviation = accountTypeAbbreviation;
		}

		public String getAccountTypeAbbreviation() {
			return accountTypeAbbreviation;
		}

		public static String valuesAsString() {
			StringBuilder builder = new StringBuilder("");
			AccountType[] values = AccountType.values();
			int i = 0;
			for (AccountType value : values) {
				boolean last = values.length - 1 == i;
				if (last) {
					builder.append(value.name());
				} else {
					builder.append(value.name()).append(", ");
				}
				i++;
			}
			return builder.toString();

		}
	}

	private static enum TransactionTypeEnum {
		AuthCaptureTxn("sale"),
		AuthorizeTxn("preAuth"),
		CaptureTxn ("postAuth"),
		ForceCaptureTxn("forceTicket"),
		CreditTxn ("credit"),
		RefundTxn("return"),
		VoidTxn("void"),
		CreateSubscription("install"),
		ModifySubscription("modify"),
		CancelSubscription("cancel");

		String transactionType;

		TransactionTypeEnum(String transactionType) {
			this.transactionType = transactionType;
		}

		public String getTransactionType() {
			return transactionType;
		}


		@Override
		public String toString() {
			return getTransactionType();
		}
	}

	public static enum CardCodeIndicator {
		PROVIDED, NOT_PROVIDED, ILLEGIBLE, NO_IMPRINT, NOT_PRESENT;

		public static String valuesAsString() {
			StringBuilder builder = new StringBuilder("");
			CardCodeIndicator[] values = CardCodeIndicator.values();
			int i = 0;
			for (CardCodeIndicator value : values) {
				boolean last = values.length - 1 == i;
				if (last) {
					builder.append(value.name());
				} else {
					builder.append(value.name()).append(", ");
				}
				i++;
			}
			return builder.toString();

		}
	}

	private static enum PayPeriod {
		day, week, month, year;

		public static String valuesAsString() {
			StringBuilder builder = new StringBuilder("");
			PayPeriod[] values = PayPeriod.values();
			int i = 0;
			for (PayPeriod value : values) {
				boolean last = values.length - 1 == i;
				if (last) {
					builder.append(value.name());
				} else {
					builder.append(value.name()).append(", ");
				}
				i++;
			}
			return builder.toString();

		}
	}

	private static enum TransactionOrigin {
		ECI, MOTO, RETAIL;

		public static String valuesAsString() {
			StringBuilder builder = new StringBuilder("");
			TransactionOrigin[] values = TransactionOrigin.values();
			int i = 0;
			for (TransactionOrigin value : values) {
				boolean last = values.length - 1 == i;
				if (last) {
					builder.append(value.name());
				} else {
					builder.append(value.name()).append(", ");
				}
				i++;
			}
			return builder.toString();

		}
	}

	public static FDGGWSApiOrderService getService() {
		if(service == null) {
			URL url = FirstDataGlobalGateway.class.getResource("./wsdl/services/order.wsdl");
			service = new FDGGWSApiOrderService(url, new QName("http://secure.linkpt.net/fdggwsapi/schemas_us/fdggwsapi", "FDGGWSApiOrderService"));
		}
		return service;
	}

	private CardCodeIndicator cardCodeIndicator;

	public CardCodeIndicator getCardCodeIndicator() {
		return cardCodeIndicator;
	}

	public void setCardCodeIndicator(CardCodeIndicator cardCodeIndicator) {
		this.cardCodeIndicator = cardCodeIndicator;
	}

	/*private void initializeSecuritySettings(String certificatePath, String certificatePassword) throws GeneralSecurityException, IOException {
		KeyManagerFactory keyManagerFactory = KeyManagerFactory.getInstance(KeyManagerFactory.getDefaultAlgorithm());
		KeyStore keyStore = KeyStore.getInstance(KeyStore.getDefaultType());

		if(certificatePath != null && certificatePassword != null) {
			log.fine("Configuring trust manager using client-side certificate");
			FileInputStream certStream = new FileInputStream(certificatePath);
			try {
				keyStore.load(certStream, certificatePassword.toCharArray());
				keyManagerFactory.init(keyStore, certificatePassword .toCharArray());
			} finally {
				certStream.close();
			}

			SSLContext sslContext = SSLContext.getInstance("SSL");
			KeyManager[] keyManagers = null;
			try {
				keyManagers = keyManagerFactory.getKeyManagers();
			} catch (IllegalStateException e) {
				String errMsg = "An error occurred while getting key managers from a factory";
				log.log(Level.WARNING, errMsg, e);
			}

			sslContext.init(keyManagers, null, null);
			HttpsURLConnection.setDefaultSSLSocketFactory(sslContext.getSocketFactory());
		} else {
			log.fine("No client-side certificate info provided, so trust manager will not be configured");
		}
	}*/

	private void createAndInsertHandlers(BindingProvider bindingProvider, String username, String password, String url) {
		bindingProvider.getRequestContext().put(BindingProvider.USERNAME_PROPERTY, username);
		bindingProvider.getRequestContext().put(BindingProvider.PASSWORD_PROPERTY, password);
		bindingProvider.getRequestContext().put(BindingProvider.ENDPOINT_ADDRESS_PROPERTY, url);
		Binding binding = bindingProvider.getBinding();
		List<Handler> handlers = binding.getHandlerChain();
		loggingHandler = new SOAPLoggingHandler() {
			@Override
			public void setLastInboundMsg(String lastInboundMsg) {
				super.setLastInboundMsg(lastInboundMsg);
			}
		};
		handlers.add(loggingHandler);

		soapFaultEditingHandler = new SoapFaultEditingHandler() {
			@Override
			public void setLastFaultMessage(String lastFaultMessage) {
				super.setLastFaultMessage(lastFaultMessage);
			}
		};
		handlers.add(soapFaultEditingHandler);
		binding.setHandlerChain(handlers);
	}

	private void buildTransaction(Transaction transaction, TransactionTypeEnum transactionType, Order order) throws PaymentAuthorizationException {
		MethodOfPayment methodOfPayment = order.getMethodOfPayment();

		CreditCard creditCard;

		String transactionOrigin = ((FirstDataOrder) order).getTransactionOrigin();
		if(methodOfPayment instanceof CreditCard) {
			CreditCardTxType txType = new CreditCardTxType();
			txType.setType(transactionType.getTransactionType());
			transaction.setCreditCardTxType(txType);
			CreditCardData cardData = new CreditCardData();
			creditCard = ((CreditCard) methodOfPayment);

			String trackData = null;

			if(creditCard.getSwipe() != null) {
				trackData = creditCard.getSwipe();
			} else if(creditCard.getTrack1() != null) {
				trackData = creditCard.getTrack1();
			} else if(creditCard.getTrack2() != null) {
				trackData = creditCard.getTrack2();
			}

			if(trackData != null && creditCard.isCardPresent()) {
				cardData.setTrackData(trackData);
				transactionOrigin = TransactionOrigin.RETAIL.name();
			} else {
				cardData.setCardNumber(creditCard.getCardNumber());
				if (creditCard.getExpirationDate() == null) {
					throw new PaymentAuthorizationException("You must specify an expiration date when processing a transaction that does not include track (magnetic card swipe) data");
				}
				cardData.setExpMonth(monthFormat.format(creditCard.getExpirationDate()));
				cardData.setExpYear(yearFormat.format(creditCard.getExpirationDate()));
				if(getCardCodeIndicator() != null) {
					cardData.setCardCodeIndicator(getCardCodeIndicator().name());
				}
				if(creditCard.getCardVerificationValue() != null) {
					cardData.setCardCodeIndicator(CardCodeIndicator.PROVIDED.name());
					cardData.setCardCodeValue(creditCard.getCardVerificationValue());
				}
			}
			transaction.setCreditCardData(cardData);
		} else if(methodOfPayment instanceof BankAccount) {
			TeleCheckTxType txType = new TeleCheckTxType();
			txType.setType(transactionType.getTransactionType());
			transaction.setTeleCheckTxType(txType);

			TeleCheckData checkData = new TeleCheckData();
			BankAccount bankAccount = (BankAccount) methodOfPayment;

			checkData.setCheckNumber(bankAccount.getCheckNumber());
			String accountType = bankAccount.getAccountType();
			if(accountType != null) {
				try {
					checkData.setAccountType(AccountType.valueOf(accountType.toUpperCase()).getAccountTypeAbbreviation());
				} catch (IllegalArgumentException e) {
					String errMsg = "Invalid bank account type. Valid values are " + AccountType.valuesAsString() + ".";
					log.log(Level.SEVERE, errMsg, e);
					throw new PaymentAuthorizationException(errMsg);
				}
			}
			checkData.setAccountNumber(bankAccount.getAccountNumber());
			checkData.setRoutingNumber(bankAccount.getAbaCode());
			checkData.setDrivingLicenseNumber(bankAccount.getDriversLicenseNumber());
			checkData.setDrivingLicenseState(bankAccount.getDriversLicenseState());
			transaction.setTeleCheckData(checkData);
		}

		Payment payment = new Payment();
		payment.setChargeTotal(order.getAmount());

		transaction.setPayment(payment);

		TransactionDetails transactionDetails = new TransactionDetails();

		transactionDetails.setInvoiceNumber(order.getInvoiceNumber());
		transactionDetails.setOrderId(order.getOrderId());
		transactionDetails.setReferenceNumber(order.getReferenceNumber());
		try {
			transactionDetails.setTransactionOrigin(transactionOrigin == null ? null : TransactionOrigin.valueOf(transactionOrigin.toUpperCase()).name());
		} catch (IllegalArgumentException e) {
			String errMsg = "Invalid transaction origin specified. Valid values are " + TransactionOrigin.valuesAsString();
			log.log(Level.SEVERE, errMsg, e);
			throw new PaymentAuthorizationException(errMsg);
		}
		transactionDetails.setPONumber(order.getPoNumber());

		transaction.setTransactionDetails(transactionDetails);

		Billing billing = new Billing();

		Customer customer = order.getCustomer();
		billing.setCustomerID(customer.getCustomerId());
		Address billingAddress = customer.getBillingAddress();

		billing.setName(GatewayUtils.getFullName(billingAddress.getFirstName(), billingAddress.getLastName()));

		billing.setCompany(billingAddress.getCompanyName());
		billing.setAddress1(billingAddress.getAddress());
		billing.setAddress2(billingAddress.getAddress2());
		billing.setCity(billingAddress.getCity());
		billing.setState(billingAddress.getState());
		billing.setZip(billingAddress.getZip());
		billing.setCountry(billingAddress.getCountry());
		billing.setPhone(billingAddress.getPhone());
		billing.setFax(billingAddress.getFax());
		billing.setEmail(billingAddress.getEmail());

		transaction.setBilling(billing);

		Shipping shipping = new Shipping();

		Address shippingAddress = order.getCustomer().getShippingAddress();
		shipping.setName(GatewayUtils.getFullName(shippingAddress.getFirstName(), shippingAddress.getLastName()));
		shipping.setAddress1(shippingAddress.getAddress());
		shipping.setAddress2(shippingAddress.getAddress2());
		shipping.setCity(shippingAddress.getCity());
		shipping.setState(shippingAddress.getState());
		shipping.setZip(shippingAddress.getZip());
		shipping.setCountry(shippingAddress.getCountry());

		transaction.setShipping(shipping);

		Notes notes = new Notes();
		notes.setComments(order.getDescription());
		transaction.setNotes(notes);
	}

	private void buildRecurringPayment(RecurringPayment recurringPayment, TransactionTypeEnum transactionType, Subscription subscription) throws PaymentAuthorizationException {
		Order order = subscription.getOrder();

		MethodOfPayment methodOfPayment = order.getMethodOfPayment();
		CreditCard creditCard;

		recurringPayment.setFunction(transactionType.getTransactionType());

		RecurringPaymentInformation recurringPaymentInformation = new RecurringPaymentInformation();
		recurringPaymentInformation.setRecurringStartDate(recurringFormat.format(subscription.getSubscriptionStartDate()));
		recurringPaymentInformation.setInstallmentCount(subscription.getNumberOfInstallments());
		recurringPaymentInformation.setInstallmentFrequency(BigInteger.valueOf(subscription.getFrequency()));
		try {
			recurringPaymentInformation.setInstallmentPeriod(subscription.getPeriod() == null ? null : PayPeriod.valueOf(subscription.getPeriod().toLowerCase()).name());
		} catch (IllegalArgumentException e) {
			String errMsg = "Invalid subscription installment period specified. Valid values are " + PayPeriod.valuesAsString();
			log.log(Level.SEVERE, errMsg, e);
			throw new PaymentAuthorizationException(errMsg);
		}
		recurringPaymentInformation.setMaximumFailures(subscription.getMaxFailures());

		recurringPayment.setRecurringPaymentInformation(recurringPaymentInformation);

		RecurringPayment.TransactionDataType transaction = new RecurringPayment.TransactionDataType();
		String transactionOrigin = ((FirstDataOrder) order).getTransactionOrigin();
		if(methodOfPayment instanceof CreditCard) {
			CreditCardTxType txType = new CreditCardTxType();
			txType.setType(transactionType.getTransactionType());
			CreditCardData cardData = new CreditCardData();
			creditCard = ((CreditCard) methodOfPayment);

			String trackData = null;

			if(creditCard.getSwipe() != null) {
				trackData = creditCard.getSwipe();
			} else if(creditCard.getTrack1() != null) {
				trackData = creditCard.getTrack1();
			} else if(creditCard.getTrack2() != null) {
				trackData = creditCard.getTrack2();
			}

			if(trackData != null && creditCard.isCardPresent()) {
				cardData.setTrackData(trackData);
				transactionOrigin = TransactionOrigin.RETAIL.name();
			} else {
				cardData.setCardNumber(creditCard.getCardNumber());
				cardData.setExpMonth(monthFormat.format(creditCard.getExpirationDate()));
				cardData.setExpYear(yearFormat.format(creditCard.getExpirationDate()));
			}
			if(getCardCodeIndicator() != null) {
				cardData.setCardCodeIndicator(getCardCodeIndicator().name());
			}
			cardData.setCardCodeValue(creditCard.getCardVerificationValue());

			transaction.setCreditCardData(cardData);
		} else if(methodOfPayment instanceof BankAccount) {
			TeleCheckTxType txType = new TeleCheckTxType();
			txType.setType(transactionType.getTransactionType());

			TeleCheckData checkData = new TeleCheckData();
			BankAccount bankAccount = (BankAccount) methodOfPayment;

			checkData.setCheckNumber(bankAccount.getCheckNumber());
			checkData.setAccountType(bankAccount.getAccountType());
			checkData.setAccountNumber(bankAccount.getAccountNumber());
			checkData.setRoutingNumber(bankAccount.getAbaCode());
			checkData.setDrivingLicenseNumber(bankAccount.getDriversLicenseNumber());
			checkData.setDrivingLicenseState(bankAccount.getDriversLicenseState());
			transaction.setTeleCheckData(checkData);
		}
		recurringPayment.setTransactionDataType(transaction);

		Payment payment = new Payment();
		payment.setChargeTotal(order.getAmount());

		recurringPayment.setPayment(payment);

		TransactionDetails transactionDetails = new TransactionDetails();

		transactionDetails.setInvoiceNumber(order.getInvoiceNumber());
		transactionDetails.setOrderId(order.getOrderId());
		transactionDetails.setReferenceNumber(order.getReferenceNumber());
		try {
			transactionDetails.setTransactionOrigin(transactionOrigin == null ? null : TransactionOrigin.valueOf(transactionOrigin.toUpperCase()).name());
		} catch (IllegalArgumentException e) {
			String errMsg = "Invalid transaction origin specified. Valid values are " + TransactionOrigin.valuesAsString();
			log.log(Level.SEVERE, errMsg, e);
			throw new PaymentAuthorizationException(errMsg);
		}
		transactionDetails.setPONumber(order.getPoNumber());

		recurringPayment.setTransactionDetails(transactionDetails);

		Billing billing = new Billing();

		Customer customer = order.getCustomer();
		billing.setCustomerID(customer.getCustomerId());
		Address billingAddress = customer.getBillingAddress();

		billing.setName(GatewayUtils.getFullName(billingAddress.getFirstName(), customer.getBillingAddress().getLastName()));

		billing.setCompany(billingAddress.getCompanyName());
		billing.setAddress1(billingAddress.getAddress());
		billing.setAddress2(billingAddress.getAddress2());
		billing.setCity(billingAddress.getCity());
		billing.setState(billingAddress.getState());
		billing.setZip(billingAddress.getZip());
		billing.setCountry(billingAddress.getCountry());
		billing.setPhone(billingAddress.getPhone());
		billing.setFax(billingAddress.getFax());
		billing.setEmail(billingAddress.getEmail());

		recurringPayment.setBilling(billing);

		Shipping shipping = new Shipping();

		Address shippingAddress = order.getCustomer().getShippingAddress();
		shipping.setAddress1(shippingAddress.getAddress());
		shipping.setAddress2(shippingAddress.getAddress2());
		shipping.setCity(shippingAddress.getCity());
		shipping.setState(shippingAddress.getState());
		shipping.setZip(shippingAddress.getZip());
		shipping.setCountry(shippingAddress.getCountry());

		recurringPayment.setShipping(shipping);

		Notes notes = new Notes();
		notes.setComments(order.getDescription());
		recurringPayment.setNotes(notes);
	}

	@Override
	public ProfileSupport getProfileSupport() {
		throw new UnsupportedOperationException("This operation is not supported");
	}

	@Override
	public SinglePaymentSupport getSinglePaymentSupport() {
		return new SinglePaymentSupport() {

			public void authCaptureTxn(GatewayAuthentication auth, Order order) throws PaymentAuthorizationException, PaymentIOException {
				FDGGWSApiOrder port = init(auth);

				FDGGWSApiOrderRequest orderRequest = new FDGGWSApiOrderRequest();

				Transaction transaction = new Transaction();

				transactionType = TransactionTypeEnum.AuthCaptureTxn;
				buildTransaction(transaction, transactionType, order);

				orderRequest.setTransaction(transaction);

				try {
					orderResponse = port.fdggwsApiOrder(orderRequest);
				} catch (SOAPFaultException e) {
					String lastResponse = loggingHandler.getLastInboundMsg();
					String errMsg = "A soap fault occurred while trying to perform a transaction of type " + transactionType;
					log.log(Level.SEVERE, errMsg, e);

					throw new PaymentAuthorizationException(lastResponse != null ? lastResponse : errMsg);
				}
				if(!orderResponse.getTransactionResult().equals("APPROVED")) {
					throw new PaymentAuthorizationException(orderResponse.getErrorMessage(), loggingHandler.getLastInboundMsg());
				}
			}

			public void authorizeTxn(GatewayAuthentication auth, Order order) throws PaymentAuthorizationException, PaymentIOException {
				FDGGWSApiOrder port = init(auth);

				FDGGWSApiOrderRequest orderRequest = new FDGGWSApiOrderRequest();

				Transaction transaction = new Transaction();

				transactionType = TransactionTypeEnum.AuthorizeTxn;
				buildTransaction(transaction, transactionType, order);

				orderRequest.setTransaction(transaction);

				try {
					orderResponse = port.fdggwsApiOrder(orderRequest);
				} catch (SOAPFaultException e) {
					String lastResponse = loggingHandler.getLastInboundMsg();
					String errMsg = "A soap fault occurred while trying to perform a transaction of type " + transactionType;
					log.log(Level.SEVERE, errMsg, e);

					throw new PaymentAuthorizationException(lastResponse != null ? lastResponse : errMsg);
				}
				if(!orderResponse.getTransactionResult().equals("APPROVED")) {
					throw new PaymentAuthorizationException(orderResponse.getErrorMessage(), loggingHandler.getLastInboundMsg());
				}
			}

			public void captureTxn(GatewayAuthentication auth, Order order, PreviousTransactionIdentifier previous) throws PaymentAuthorizationException, PaymentIOException {
				FDGGWSApiOrder port = init(auth);
				FDGGWSApiOrderRequest orderRequest = new FDGGWSApiOrderRequest();

				Transaction transaction = new Transaction();
				CreditCardTxType txType = new CreditCardTxType();

				transactionType = TransactionTypeEnum.CaptureTxn;
				String transactionTypeString = transactionType.getTransactionType();
				txType.setType(transactionTypeString);
				transaction.setCreditCardTxType(txType);

				Payment payment = new Payment();
				if(order.getAmount() != null) {
					order.setAmount(order.getAmount());
				}

				transaction.setPayment(payment);

				TransactionDetails transactionDetails = new TransactionDetails();
				transactionDetails.setOrderId(order.getOrderId());

				transaction.setTransactionDetails(transactionDetails);

				orderRequest.setTransaction(transaction);

				try {
					orderResponse = port.fdggwsApiOrder(orderRequest);
				} catch (SOAPFaultException e) {
					String lastResponse = loggingHandler.getLastInboundMsg();
					String errMsg = "A soap fault occurred while trying to perform a transaction of type " + transactionTypeString;
					log.log(Level.SEVERE, errMsg, e);

					throw new PaymentAuthorizationException(lastResponse != null ? lastResponse : errMsg);
				}
				if(!orderResponse.getTransactionResult().equals("APPROVED")) {
					throw new PaymentAuthorizationException(orderResponse.getErrorMessage(), loggingHandler.getLastInboundMsg());
				}
			}

			public void voidTxn(GatewayAuthentication auth, Order order, PreviousOrderIdentifier previous) throws PaymentAuthorizationException, PaymentIOException {
				FDGGWSApiOrder port = init(auth);

				FDGGWSApiOrderRequest orderRequest = new FDGGWSApiOrderRequest();

				Transaction transaction = new Transaction();
				CreditCardTxType txType = new CreditCardTxType();

				transactionType = TransactionTypeEnum.VoidTxn;
				String transactionTypeString = transactionType.getTransactionType();
				txType.setType(transactionTypeString);
				transaction.setCreditCardTxType(txType);

				TransactionDetails transactionDetails = new TransactionDetails();
				transactionDetails.setOrderId(previous.getPreviousOrderId());
				transactionDetails.setTDate(((FirstDataPreviousOrderIdentifier) previous).gettDate());

				transaction.setTransactionDetails(transactionDetails);

				orderRequest.setTransaction(transaction);

				try {
					orderResponse = port.fdggwsApiOrder(orderRequest);
				} catch (SOAPFaultException e) {
					String lastResponse = loggingHandler.getLastInboundMsg();
					String errMsg = "A soap fault occurred while trying to perform a transaction of type " + transactionTypeString;
					log.log(Level.SEVERE, errMsg, e);

					throw new PaymentAuthorizationException(lastResponse != null ? lastResponse : errMsg);
				}
				if(!orderResponse.getTransactionResult().equals("APPROVED")) {
					throw new PaymentAuthorizationException(orderResponse.getErrorMessage(), loggingHandler.getLastInboundMsg());
				}
			}

			public void refundTxn(GatewayAuthentication auth, Order order, PreviousOrderIdentifier previous) throws PaymentAuthorizationException, PaymentIOException {
				FDGGWSApiOrder port = init(auth);

				FDGGWSApiOrderRequest orderRequest = new FDGGWSApiOrderRequest();

				Transaction transaction = new Transaction();

				transactionType = TransactionTypeEnum.RefundTxn;

				if(order.getOrderId() == null) {
					transactionType = TransactionTypeEnum.CreditTxn;
					buildTransaction(transaction, transactionType, order);
				} else if(order.getMethodOfPayment() instanceof  CreditCard){
					CreditCardTxType txType = new CreditCardTxType();
					txType.setType(transactionType.getTransactionType());
					transaction.setCreditCardTxType(txType);
				} else if(order.getMethodOfPayment() instanceof BankAccount) {
					TeleCheckTxType txType = new TeleCheckTxType();
					txType.setType(transactionType.getTransactionType());
					transaction.setTeleCheckTxType(txType);
				}

				Payment payment = new Payment();
				payment.setChargeTotal(order.getAmount());
				transaction.setPayment(payment);

				TransactionDetails transactionDetails = new TransactionDetails();
				transactionDetails.setOrderId(previous.getPreviousOrderId());
				transactionDetails.setTDate(((FirstDataPreviousOrderIdentifier) previous).gettDate());
				transaction.setTransactionDetails(transactionDetails);

				orderRequest.setTransaction(transaction);

				try {
					orderResponse = port.fdggwsApiOrder(orderRequest);
				} catch (SOAPFaultException e) {
					String lastResponse = loggingHandler.getLastInboundMsg();
					String errMsg = "A soap fault occurred while trying to perform a transaction of type " + transactionType;
					log.log(Level.SEVERE, errMsg, e);

					throw new PaymentAuthorizationException(lastResponse != null ? lastResponse : errMsg);
				}
				if(!orderResponse.getTransactionResult().equals("APPROVED")) {
					throw new PaymentAuthorizationException(orderResponse.getErrorMessage(), loggingHandler.getLastInboundMsg());
				}
			}
		};
	}

	@Override
	public PaymentResponse getPaymentResponse() {
		return new CreditCardResponse() {
			public String getAvsResponse() {
				if(isSubscription()) {
					return actionResponse.getAVSResponse();
				} else {
					return orderResponse.getAVSResponse();
				}
			}

			public String getCvvResponse() {
				throw new UnsupportedOperationException("Credit-card verification responses are not supported by this gateway");
			}

			public String getTransactionId() {
				if(isSubscription()) {
					return actionResponse.getOrderId();
				} else {
					return orderResponse.getOrderId() + "|" + orderResponse.getTDate();
				}
			}

			public String getTransactionResult() {
				if(isSubscription()) {
					return actionResponse.getTransactionResult();
				} else {
					return orderResponse.getTransactionResult();
				}
			}

			public String getRawResponse() {
				return loggingHandler.getLastInboundMsg();
			}

			public String getAuthorizationCode() {
				if(isSubscription()) {
					return actionResponse.getApprovalCode();
				} else {
					return orderResponse.getApprovalCode();
				}
			}
		};
	}

	private boolean isSubscription() {
		return transactionType.equals(TransactionTypeEnum.ModifySubscription) ||
				transactionType.equals(TransactionTypeEnum.CancelSubscription) ||
				transactionType.equals(TransactionTypeEnum.CreateSubscription);
	}

	@Override
	public RecurringPaymentSupport getRecurringPaymentSupport() {
		return new RecurringPaymentSupport() {
			public void createSubscription(GatewayAuthentication auth, Subscription subscription) throws PaymentAuthorizationException, PaymentIOException {
				FDGGWSApiOrder port = init(auth);
				FDGGWSApiActionRequest request = new FDGGWSApiActionRequest();

				Action action = new Action();
				RecurringPayment recurringPayment = new RecurringPayment();

				transactionType = TransactionTypeEnum.CreateSubscription;
				buildRecurringPayment(recurringPayment, transactionType, subscription);

				action.setRecurringPayment(recurringPayment);

				request.setAction(action);

				try {
					actionResponse = port.fdggwsApiAction(request);
				} catch (SOAPFaultException e) {
					String lastResponse = loggingHandler.getLastInboundMsg();
					String errMsg = "A soap fault occurred while trying to perform a transaction of type " + transactionType;
					log.log(Level.SEVERE, errMsg, e);

					throw new PaymentAuthorizationException(lastResponse != null ? lastResponse : errMsg);
				}
				if(!actionResponse.getTransactionResult().equals("APPROVED")) {
					throw new PaymentAuthorizationException(actionResponse.getErrorMessage(), loggingHandler.getLastInboundMsg());
				}

			}

			public void updateSubscription(GatewayAuthentication auth, Subscription subscription, PreviousTransactionIdentifier previous) throws PaymentAuthorizationException, PaymentIOException {
				FDGGWSApiOrder port = init(auth);
				FDGGWSApiActionRequest request = new FDGGWSApiActionRequest();

				Action action = new Action();
				RecurringPayment recurringPayment = new RecurringPayment();
				recurringPayment.setOrderId(((PreviousOrderIdentifier) previous).getPreviousOrderId());

				transactionType = TransactionTypeEnum.ModifySubscription;
				buildRecurringPayment(recurringPayment, transactionType, subscription);

				action.setRecurringPayment(recurringPayment);

				request.setAction(action);

				try {
					actionResponse = port.fdggwsApiAction(request);
				} catch (SOAPFaultException e) {
					String lastResponse = loggingHandler.getLastInboundMsg();
					String errMsg = "A soap fault occurred while trying to perform a transaction of type " + transactionType;
					log.log(Level.SEVERE, errMsg, e);

					throw new PaymentAuthorizationException(lastResponse != null ? lastResponse : errMsg);
				}
				if(!actionResponse.getTransactionResult().equals("APPROVED")) {
					throw new PaymentAuthorizationException(actionResponse.getErrorMessage(), loggingHandler.getLastInboundMsg());
				}
			}

			public void cancelSubscription(GatewayAuthentication auth, Subscription subscription, PreviousTransactionIdentifier previous) throws PaymentAuthorizationException {
				FDGGWSApiOrder port = init(auth);
				FDGGWSApiActionRequest request = new FDGGWSApiActionRequest();

				Action action = new Action();
				RecurringPayment recurringPayment = new RecurringPayment();
				recurringPayment.setOrderId(((PreviousOrderIdentifier) previous).getPreviousOrderId());
				transactionType = TransactionTypeEnum.CancelSubscription;
				recurringPayment.setFunction(transactionType.getTransactionType());

				action.setRecurringPayment(recurringPayment);

				request.setAction(action);

				try {
					actionResponse = port.fdggwsApiAction(request);
				} catch (SOAPFaultException e) {
					String lastResponse = loggingHandler.getLastInboundMsg();
					String errMsg = "A soap fault occurred while trying to perform a transaction of type " + transactionType;
					log.log(Level.SEVERE, errMsg, e);

					throw new PaymentAuthorizationException(lastResponse != null ? lastResponse : errMsg);
				}
				if(!actionResponse.getTransactionResult().equals("APPROVED")) {
					throw new PaymentAuthorizationException(actionResponse.getErrorMessage(), loggingHandler.getLastInboundMsg());
				}
			}

		};
	}

	private FDGGWSApiOrder init(GatewayAuthentication auth) throws PaymentAuthorizationException {
		try {
			Certificate certificate = auth.getCertificate();
			GatewayUtils.addClientCertificateToKeyStore(certificate.getCertificatePath(), certificate.getCertificatePassword());
		} catch (Exception e) {
			String errMsg = "An error occurred while initializing the client-side certificate settings";
			log.log(Level.SEVERE, errMsg, e);
			throw new PaymentAuthorizationException(errMsg);
		}

		FDGGWSApiOrder port = getService().getFDGGWSApiOrderPort();

		String url;
		if(auth.getEndpointUrl() != null) {
			url = auth.getEndpointUrl();
		} else if (isTestMode()) {
			url = TEST_URL;
		} else {
			url = PRODUCTION_URL;
		}

		createAndInsertHandlers((BindingProvider) port, auth.getApiUsername(), auth.getApiPassword(), url);
		return port;
	}
}
